// $Id: cheops.cpp,v 1.6 2003/12/07 19:43:33 psy Exp $
//
// cheops.cpp : Defines the entry point for the console application.
//
// Copyright (C) 2003 Tristan Miller <psychonaut@nothingisreal.com>
//  
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of the
// License, or (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
// 02111-1307, USA.
//
//////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "stdafx.h"
#include "ChessBoard.h"
#include "HumanPlayer.h"
#include "ComputerPlayer.h"
#include <cstdio>
#include <cstring>
#include <string>
#include <sstream>
#include <vector>

void show_menu(void);
move *parse_move(const string& s);
void who_plays(Player *&w, Player *&b);
void show_help(void);
void show_version(void);

int main(int argc, char* argv[]) {

    ChessBoard c;
    Player *w_player=NULL, *b_player=NULL, *this_player;
    ComputerPlayer *computer_player=NULL;
    string cmd, log;
    game_status status=Normal;
    move *m = NULL;
    int turn=0;
    
    bool uci = false;

    // Process command-line parameters
    while (argc-- > 1) {
      argv++;
      if (strcmp(*argv, "--help") == 0) {
	show_help();
	exit(0);
      }
      else if (strcmp(*argv, "--version") == 0) {
	show_version();
	exit(0);
      }
      else if (strcmp(*argv, "--uci") == 0) {
          uci = true;
      }
      else {
	cerr << "cheops: unrecognized option `" << argv[0] << "'\n";
	show_help();
	exit(1);
      }
    }
    
    // Print title screen
    show_version();
    cout <<
      "\t\t\t__   __\n"
      "\t\t\t\\ \\ / /\n"
      "\t\t\t \\ v / ___ ___  _  _  _\n"
      "\t\t\t  > < / __) _ \\| || || |\n"
      "\t\t\t / ^ \\> _| (_) ) \\| |/ |\n"
      "\t\t\t/_/ \\_\\___)___/ \\_   _/\n"
      "\t\t\t _________________| |__\n"
      "\t\t\t(_________________|_|__)\n\n"
      ;
    
    if (uci)
    {
       // initialize computer player
       
       computer_player = new ComputerPlayer(true);
       
       ComputerPlayer::show_thinking = true;
       
       // initialize chess board
       
       c = ChessBoard();
       
       // wait for and process next UCI command
       // tokenizer from : http://oopweb.com/CPP/Documents/CPPHOWTO/Volume/C++Programming-HOWTO-7.html

       string str;                // string to hold UCI command
       string buf;                // string buffer
       string move_str;           // computer move string
       vector<string> tokens;     // vector to hold tokens
       
       const int NUM_PARAMS = 34;
       string option[34];
       
       option[ 0] = "draw_value";
       option[ 1] = "endgame_threshold";
       option[ 2] = "pawn_material";
       option[ 3] = "doubled_pawn";
       option[ 4] = "backward_pawn";
       option[ 5] = "pawn_proximity_to_opposite_edge";
       option[ 6] = "isolated_pawn_on_file_1_or_8";
       option[ 7] = "isolated_pawn_on_file_2_or_7";
       option[ 8] = "isolated_pawn_on_file_3_or_6";
       option[ 9] = "isolated_pawn_on_file_4_or_5";
       option[10] = "knight_material";
       option[11] = "knight_distance_to_either_king";
       option[12] = "knight_proximity_to_centre";
       option[13] = "bishop_material";
       option[14] = "bishop_proximity_to_centre";
       option[15] = "bishop_mobility";
       option[16] = "bishop_X-ray_mobility";
       option[17] = "presence_of_both_bishops";
       option[18] = "bishop_mobility_bonus";
       option[19] = "bishop_X-ray_mobility_bonus";
       option[20] = "rook_material";
       option[21] = "rook_proximity_to_enemy_king";
       option[22] = "rook_mobility_type";
       option[23] = "rook_X-ray_mobility";
       option[24] = "friendly_pawns_on_same_file_as_rook";
       option[25] = "rook_mobility_bonus";
       option[26] = "rook_X-ray_mobility_bonus";
       option[27] = "queen_material";
       option[28] = "queen_proximity_to_enemy_king";
       option[29] = "queen_proximity_to_centre";
       option[30] = "king_distance_from_centre_(except_endgame)";
       option[31] = "no_friendly_pawns_adjacent_to_king";
       option[32] = "king_castling";
       option[33] = "search_ply";
              
       while (1)
       {
          getline(cin, str);
          stringstream ss(str);
          
          // tokenize string
          
          tokens.clear();

          while (ss >> buf)
          {
             tokens.push_back(buf);
          }
          
          /*
          for (int i = 0; i < tokens.size(); i++)
          {
             cout << "[" << tokens[0] << "] ";
          }
          cout << endl;
          */
          
          // make sure at least one element
          
          if (tokens.size() >= 1)
          {
             cmd = tokens[0];
          }
          else
          {
             continue;
          }
          
          if (cmd == "go" || cmd == "ponderhit")
          {
             move_str = computer_player->get_command(c);
             
             delete m;
             
             /*
             if (move_str == "o-o")
             {
                m = new move;
                m->promotion = Empty;
                m->from = (turn % 2 == 0 ? e1 : e8);
                m->to = (turn % 2 == 0 ? g1 : g8);
             }
             else if (move_str == "o-o-o")
             {
                m = new move;
                m->promotion = Empty;
                m->from = (turn % 2 == 0 ? e1 : e8);
                m->to = (turn % 2 == 0 ? c1 : c8);
             }
             else
             {
                m = parse_move(move_str);
             }
             */
             
             m = parse_move(move_str);
          }
          else if (cmd == "isready")
          {
             // synchronize with GUI
             
             cout << "readyok\n";
          }
          else if (cmd == "ponderhit")
          {
             // no pondering anyway... so just respond as if got go command (see above)
          }
          else if (cmd == "position")
          {
             // set current position
             
             if (tokens.size() >= 2)
             {
                if (tokens[1] == "fen")
                {
                   cout << "WARNING : cheops doesn't support the position fen command yet\n";                
                }
                else if (tokens[1] == "startpos")
                {
                   // put board in start position
                   
                   c = ChessBoard();
                   turn = 0;
                   
                   // see if move list
                   
                   if (tokens.size() >= 4 && tokens[2] == "moves")
                   {
                      // "position startpos moves e2e4 e7e5 g1f3 ..."
                      
                      for (int i = 3; i < tokens.size(); i++)
                      {
                         delete m;
                         m = parse_move(tokens[i]);                         
                         status=c.do_move(*m);
                         turn++;
                      }
                   }
                }
             }
             else
             {
                cout << "ERROR : not enough arguments for position command\n";
             }
          }
          else if (cmd == "setoption")
          {
             // set value of option : "setoption name knight_material value 300"
             
             if (tokens.size() != 5)
             {
                cout << "ERROR : incorrect number of arguments for setoption command\n";
                continue;
             }
             
             if (tokens[1] != "name" || tokens[3] != "value")
             {
                cout << "ERROR : incorrect form of setoption command\n";
                continue;
             }
             
             // make sure param exists

             int paramIndex = -1;
             for (int i = 0; i < NUM_PARAMS; i++)
             {
                if (option[i] == tokens[2])
                {
                   paramIndex = i;
                   break;
                }
             }
             
             if (paramIndex == -1)
             {
                cout << "ERROR : unknown UCI parameter [" << tokens[2] << "]\n";
                continue;
             }
             
             computer_player->set_single_stat(paramIndex, atol(tokens[4].c_str()));
          }
          else if (cmd == "stop")
          {
             // ignore for now since not interrupting search
          }
          else if (cmd == "quit")
          {
             // exit immediately
             
             break;
          }
          else if (cmd == "uci")
          {
             // send id and possible options
             
             cout << "id name CHEOPS 1.1\n";
             cout << "id author Tristan Miller\n";
             
             cout << "option name draw_value type spin default 0 min -100 max 100\n";
             cout << "option name endgame_threshold type spin default 1500 min 0 max 3000\n";
             cout << "option name pawn_material type spin default 100 min 0 max 1200\n";
             cout << "option name doubled_pawn type spin default -12 min -100 max 100\n";
             cout << "option name backward_pawn type spin default -6 min -100 max 100\n";
             cout << "option name pawn_proximity_to_opposite_edge type spin default 1 min -100 max 100\n";
             cout << "option name isolated_pawn_on_file_1_or_8 type spin default -12 min -100 max 100\n";
             cout << "option name isolated_pawn_on_file_2_or_7 type spin default -14 min -100 max 100\n";
             cout << "option name isolated_pawn_on_file_3_or_6 type spin default -16 min -100 max 100\n";
             cout << "option name isolated_pawn_on_file_4_or_5 type spin default -20 min -100 max 100\n";
             cout << "option name knight_material type spin default 330 min 0 max 1200\n";
             cout << "option name knight_distance_to_either_king type spin default -1 min -100 max 100\n";
             cout << "option name knight_proximity_to_centre type spin default 10 min -100 max 100\n";
             cout << "option name bishop_material type spin default 330 min 0 max 1200\n";
             cout << "option name bishop_proximity_to_centre type spin default 10 min -100 max 100\n";
             cout << "option name bishop_mobility type spin default 1 min -100 max 100\n";
             cout << "option name bishop_X-ray_mobility type spin default 1 min -100 max 100\n";
             cout << "option name presence_of_both_bishops type spin default -20 min -100 max 100\n";
             cout << "option name bishop_mobility_bonus type spin default 8 min -100 max 100\n";
             cout << "option name bishop_X-ray_mobility_bonus type spin default 8 min -100 max 100\n";
             cout << "option name rook_material type spin default 500 min 0 max 1200\n";
             cout << "option name rook_proximity_to_enemy_king type spin default 20 min -100 max 100\n";
             cout << "option name rook_mobility_type spin default 1 min -100 max 100\n";
             cout << "option name rook_X-ray_mobility type spin default 1 min -100 max 100\n";
             cout << "option name friendly_pawns_on_same_file_as_rook type spin default 10 min -100 max 100\n";
             cout << "option name rook_mobility_bonus type spin default 8 min -100 max 100\n";
             cout << "option name rook_X-ray_mobility_bonus type spin default 8 min -100 max 100\n";
             cout << "option name queen_material type spin default 900 min 0 max 1200\n";
             cout << "option name queen_proximity_to_enemy_king type spin default 20 min -100 max 100\n";
             cout << "option name queen_proximity_to_centre type spin default 10 min -100 max 100\n";
             cout << "option name king_distance_from_centre_(except_endgame) type spin default 10 min -100 max 100\n";
             cout << "option name no_friendly_pawns_adjacent_to_king type spin default -16 min -100 max 100\n";
             cout << "option name king_castling type spin default 10 min -100 max 100\n";
             cout << "option name search_ply type spin default 3 min 1 max 20\n";  
             
             cout << "uciok\n";
          }
          else
          {
             // unknown command
             cout << "ERROR : unknown command --> [" << str << "]\n";
          }
       }
    }
    else {
        who_plays(w_player, b_player);

        cout << "\nEnter \"help\" or \"?\" for the menu.\n" << c;

        // Main loop
        while (1) {
            cout.flush();
            this_player=c.white_moves()?w_player:b_player;
            cmd=this_player->get_command(c);

            if (cmd=="board" || cmd=="bd")
                cout << c;
            else if (cmd=="reverse")
                c.show_reverse=!c.show_reverse;
            else if (cmd=="coords")
                c.show_coordinates=!c.show_coordinates;
            else if (cmd=="think")
                ComputerPlayer::show_thinking=!ComputerPlayer::show_thinking;
            else if (cmd=="help" || cmd=="?")
                show_menu();
            else if (cmd=="who")
                who_plays(w_player, b_player);
            else if (cmd=="save") {
                ofstream logfile("cheops.log");
                logfile << log << endl;
                logfile.close();
            }
            else if (cmd=="new") {
                c=ChessBoard();
                turn=0;
                log=string();
            }
            else if (cmd=="quit" || cmd=="exit")
                break;
            else if (cmd=="log")
                cout << log << endl;
            else if (cmd=="resign") {
                cout << "You resign. " << (c.white_moves()?"Black":"White") << 
                    " wins!\n";
                c.in_progress=false;
            }
            else {
                delete m;
                // Kingside castling
                if (cmd=="o-o") {
                    m = new move;
                    m->promotion=Empty;
                    m->from=this_player==w_player?e1:e8;
                    m->to=this_player==w_player?g1:g8;
                }
                // Queenside castling
                else if (cmd=="o-o-o") {
                    m = new move;
                    m->promotion=Empty;
                    m->from=this_player==w_player?e1:e8;
                    m->to=this_player==w_player?c1:c8;
                }
                // Regular move
                else
                    m=parse_move(cmd);
                if (c.in_progress && (!this_player->human || m && c.can_move(*m))) {
                    status=c.do_move(*m);
                    cout << c;

                    // Update board log
                    if (turn%2==0) {
                        char tmp[6];
                        sprintf(tmp,"%2d. ",turn/2+1);
                        log+=tmp;
                        log+=*m;
                    }
                    else {
                        log+='\t';
                        log+=*m;
                        log+='\n';
                    }
                    turn++;

                    // Display board status
                    switch (status) {
                    case Checkmate:
                        cout << "Checkmate!\n";
                        break;
                    case Stalemate:
                        cout << "Stalemate!\n";
                        break;
                    case TripleOccurrence:
                        cout << "Triple occurrence -- game is drawn.\n";
                        break;
                    case FiftyMoves:
                        cout << "Fifty move rule -- game is drawn.\n";
                    case Check:
                        cout << "Check!\n";
                        break;
                    }
                }
                else
                    cout << "Illegal move\n";
            }
        }
    }

    return 0;
}

// -------------------------------------------------------------------
// Display the menu
// -------------------------------------------------------------------
void show_menu(void) {
    cout << "\nCHEOPS command summary\n";
    cout << "---------------------------------------------------------\n";
    cout << "a1b3\tmove from a1 to b3\n";
    cout << "a7a8n\tpromote pawn to knight\n";
    cout << "o-o\tcastle king-side\n";
    cout << "o-o-o\tcastle queen-side\n";
//  cout << "draw\toffer/force draw\n";
    cout << "resign\tresign\n";
    cout << "bd\tredisplay board\n";
    cout << "coords\ttoggle coordinates\n";
    cout << "reverse\treverse board display\n";
    cout << "think\ttoggle display of computer thinking\n";
    cout << "log\tlog of moves\n";
    cout << "save\tsave log to cheops.log\n";
    cout << "new\tnew game\n";
    cout << "who\twho plays (modify current players)\n";
    cout << "quit\tquit program\n";
    cout << "---------------------------------------------------------\n";
}

// -------------------------------------------------------------------
// Parse move
//  - returns pointer to a move structure on success, else NULL
// -------------------------------------------------------------------
move *parse_move(const string& s) {

    // Is this a valid move string?
    int len=s.length();
    if (len>5 || len<4 || s[0]<'a' || s[0]>'h' || s[2]<'a' ||
        s[2]>'h' || s[1]<'1' || s[1]>'8' || s[3]<'1' || s[3]>'8')
        return NULL;
    if (len==5 && !strchr("qrnb", s[4]))
        return NULL;

    // Determine promotion type
    move *m = new move;
    if (len==5) {
        switch(s[4]) {
        case 'q':
            m->promotion=Queen;
            break;
        case 'r':
            m->promotion=Rook;
            break;
        case 'n':
            m->promotion=Knight;
            break;
        case 'b':
            m->promotion=Bishop;
            break;
        }
    }
    else 
        m->promotion=Empty;

    // Determine source and destination squares
    m->from=s[0]-'a'+(s[1]-'1')*8;
    m->to=s[2]-'a'+(s[3]-'1')*8;

    return m;
}

// -------------------------------------------------------------------
// Set players
// -------------------------------------------------------------------
void who_plays(Player *&w, Player *&b) {
    string s;
    while (s!="c" && s!="h") {
        cout << "White [c]omputer or [h]uman? ";
        cin >> s;
        cout << endl;
    }
    if (s=="h") {
        if (!(w && w->human==true)) {
            delete w;
            w = new HumanPlayer;
        }
    }
    else {
        if (!(w && w->human==false)) {
            delete w;
            w = new ComputerPlayer(false);
        }
        else
            w->set_stats();
    }

    s=string();
    while (s!="c" && s!="h") {
        cout << "Black [c]omputer or [h]uman? ";
        cin >> s;
        cout << endl;
    }
    if (s=="h") {
        if (!(b && b->human==true)) {
            delete b;
            b = new HumanPlayer;
        }
    }
    else {
        if (!(b && b->human==false)) {
            delete b;
            b = new ComputerPlayer(false);
        }
        else
            b->set_stats();
    }
}

void show_help(void) {
  cerr <<
    "Usage: cheops [options]\n"
    "Options:\n"
    "--help     Display this information and exit\n"
    "--version  Display version information and exit\n"
    "--uci      Start in UCI mode\n"
    ;
}

void show_version(void) {
  cout <<
    PACKAGE_STRING "\n"
    "Copyright (C) 1999, 2000, 2003 Tristan Miller\n"
    "This is free software; see the source for copying conditions.  There is NO\n"
    "warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n"
    ;
}
